/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include "config_manager.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libconfig.h>
#include <sys/stat.h>

#include "util.h"

#define DEFAULT_WEB_PORT        8077
#define DEFAULT_WEB_INTERFACE   ""
#define DEFAULT_WEBROOT         PACKAGE_DATADIR "/webroot"
#define DEFAULT_DS485D_URL      "tcp://localhost:8442"
#define DEFAULT_DATABASE_FILE   "/var/lib/dsa/dsa.db"
#define DEFAULT_LOG_DIR         "/var/log/dsa"
#define DEFAULT_LOG_LEVEL       "INFO"

namespace dSA
{

ConfigManager* ConfigManager::m_instance = NULL;

ConfigManager* ConfigManager::createInstance(const char* config_file,
                                             const char* webroot,
                                             int port)
{
    if (m_instance == NULL)
    {
        m_instance = new ConfigManager(config_file, webroot, port);
        if (m_instance == NULL)
        {
            throw _Exception("Could not create ConfigManager");
        }
        return m_instance;
    }
    else
    {
        throw _Exception("Can't reinit, ConfigManager is already initialized");
    }
}

ConfigManager* ConfigManager::getInstance()
{
    if (m_instance == NULL)
    {
        throw _Exception("ConfigManager must be initialized first!");
    }
    return m_instance;
}

ConfigManager::ConfigManager() {}

ConfigManager::ConfigManager(const char* config_file, const char* webroot,
                             int webport)
{
    config_t config;
    config_setting_t* cfg_root;
    std::string cfg_file_opt;
    config_setting_t* setting;
    bool config_dirty = false;
    struct stat statbuf;

    const char* ds485d_url = DEFAULT_DS485D_URL;
    const char* cfg_webroot = DEFAULT_WEBROOT;
    const char* cfg_web_iface = DEFAULT_WEB_INTERFACE;
    const char* cfg_database = DEFAULT_DATABASE_FILE;
    const char* cfg_logdir = DEFAULT_LOG_DIR;
    const char* cfg_loglevel = DEFAULT_LOG_LEVEL;
    LIBCONFIG_LOOKUP_INT_ARG cfg_port = DEFAULT_WEB_PORT;

    if (config_file == NULL)
    {
        throw _Exception("No configuratin file given");
    }
    m_filename = config_file;

    // command line options override defaults and override the config file,
    // but they do not get written to the config file automatically
    if (webroot != NULL)
    {
        if (stat(webroot, &statbuf) != 0)
        {
            throw _Exception("Invalid webroot directory " +
                             std::string(webroot));
        }

        if (!S_ISDIR(statbuf.st_mode))
        {
            throw _Exception("Specified webroot " + std::string(webroot) +
                             " is not a directory");
        }
    }

    m_options.resize(CFG_MAX);

    if (config_file != NULL)
    {
        if (stat(config_file, &statbuf) != 0)
        {
            throw _Exception("Could not find configuration file " +
                             std::string(config_file));
        }

        if (!S_ISREG(statbuf.st_mode))
        {
            throw _Exception(std::string(config_file) +
                             " is not a regular file");
        }

        config_init(&config);
        if (!config_read_file(&config, config_file))
        {
            std::string err = "Error in configuration: " +
                              std::string(config_file) + ":" +
                              util::str(config_error_line(&config)) + " " +
                              config_error_text(&config);
            config_destroy(&config);
            throw _Exception(err);
        }

        cfg_root = config_root_setting(&config);

        if (!config_lookup_string(&config, "ds485d-url", &ds485d_url))
        {
            setting = config_setting_add(cfg_root, "ds485d-url",
                                         CONFIG_TYPE_STRING);
            if (setting == NULL)
            {
                setting = config_setting_get_member(cfg_root, "ds485d-url");
            }
            ds485d_url = DEFAULT_DS485D_URL;
            config_setting_set_string(setting, ds485d_url);
            config_dirty = true;
        }

        if (!config_lookup_string(&config, "database", &cfg_database))
        {
            setting = config_setting_add(cfg_root, "database",
                                         CONFIG_TYPE_STRING);
            if (setting == NULL)
            {
                setting = config_setting_get_member(cfg_root, "database");
            }
            cfg_database = DEFAULT_DATABASE_FILE;
            config_setting_set_string(setting, cfg_database);
            config_dirty = true;
        }

        if (!config_lookup_string(&config, "logdir", &cfg_logdir))
        {
            setting = config_setting_add(cfg_root, "logdir",
                                         CONFIG_TYPE_STRING);
            if (setting == NULL)
            {
                setting = config_setting_get_member(cfg_root, "logdir");
            }
            cfg_logdir = DEFAULT_LOG_DIR;
            config_setting_set_string(setting, cfg_logdir);
            config_dirty = true;
        }

        if (!config_lookup_string(&config, "loglevel", &cfg_loglevel))
        {
            setting = config_setting_add(cfg_root, "loglevel",
                                         CONFIG_TYPE_STRING);
            if (setting == NULL)
            {
                setting = config_setting_get_member(cfg_root, "loglevel");
            }
            cfg_loglevel = DEFAULT_LOG_LEVEL;
            config_setting_set_string(setting, cfg_loglevel);
            config_dirty = true;
        }

        if (webroot == NULL)
        {
            if (!config_lookup_string(&config, "webroot", &cfg_webroot))
            {
                setting = config_setting_add(cfg_root, "webroot",
                                             CONFIG_TYPE_STRING);
                if (setting == NULL)
                {
                    setting = config_setting_get_member(cfg_root, "webroot");
                }
                cfg_webroot = DEFAULT_WEBROOT;
                config_setting_set_string(setting, cfg_webroot);
                config_dirty = true;
            }
        }
        else
        {
            cfg_webroot = webroot;
        }

        if (webport <= 0)
        {
            if (!config_lookup_int(&config, "webserver-port", &cfg_port))
            {
                setting = config_setting_add(cfg_root, "webserver-port",
                                             CONFIG_TYPE_INT);
                if (setting == NULL)
                {
                    setting = config_setting_get_member(cfg_root,
                                                        "webserver-port");
                }
                cfg_port = DEFAULT_WEB_PORT;
                config_setting_set_int(setting, cfg_port);
                config_dirty = true;
            }
        }
        else
        {
            cfg_port = webport;
        }

        if (!config_lookup_string(&config, "webserver-interface",
                                 &cfg_web_iface))
        {
            cfg_web_iface = DEFAULT_WEB_INTERFACE;
        }

        cfg_file_opt = config_file;

        if (config_dirty)
        {
            int ret = config_write_file(&config, config_file);
            if (!ret)
            {
                fprintf(stderr, "Error while writing configuration file %s\n",
                        config_file);
            }
        }
    } // if (config_file != NULL)

    m_options[CFG_DS485_DAEMON_URL] =
        dynamic_cast<GenericOption *>(new Option(ds485d_url));
    m_options[CFG_WEBSERVER_PORT] =
        dynamic_cast<GenericOption *>(new IntOption(cfg_port));
    m_options[CFG_WEBSERVER_DOCUMENT_ROOT] =
        dynamic_cast<GenericOption *>(new Option(cfg_webroot));
    m_options[CFG_WEBSERVER_INTERFACE] =
        dynamic_cast<GenericOption *>(new Option(cfg_web_iface));
    m_options[CFG_DATABASE_FILE] =
            dynamic_cast<GenericOption *>(new Option(cfg_database));
    m_options[CFG_LOG_DIR] =
            dynamic_cast<GenericOption *>(new Option(cfg_logdir));
    m_options[CFG_LOG_LEVEL] =
            dynamic_cast<GenericOption *>(new Option(cfg_loglevel));
    m_options[CFG_CONFIG_FILE] =
            dynamic_cast<GenericOption *>(new Option(cfg_file_opt));

    if (config_file != NULL)
    {
        config_destroy(&config);
    }
}

const std::string& ConfigManager::getOption(ConfigOptions option) const
{
    if ((option < 0) || (option >= CFG_MAX))
    {
        throw _Exception("Invalid option given!");
    }

    return m_options.at(option)->getOption();
}

const int ConfigManager::getIntOption(ConfigOptions option) const
{
    if ((option < 0) || (option >= CFG_MAX))
    {
        throw _Exception("Invalid option given!");
    }

    return m_options.at(option)->getIntOption();
}

const bool ConfigManager::getBoolOption(ConfigOptions option) const
{
    if ((option < 0) || (option >= CFG_MAX))
    {
        throw _Exception("Invalid option given!");
    }

    return m_options.at(option)->getBoolOption();
}

boost::shared_ptr<GenericOption> ConfigManager::readParameter(std::string path)
{
    config_t config;
    config_setting_t* setting;
    boost::shared_ptr<GenericOption> option;

    config_init(&config);
    if (!config_read_file(&config, m_filename.c_str()))
    {
        std::string err = "Error in configuration: " + m_filename + ":" +
                          util::str(config_error_line(&config)) + " " +
                          config_error_text(&config);
        config_destroy(&config);
        throw _Exception(err);
    }

    setting = config_lookup(&config, path.c_str());
    if (setting)
    {
        switch (config_setting_type(setting)) {
            case CONFIG_TYPE_INT:
                option = boost::shared_ptr<GenericOption> (new IntOption(setting->value.ival));
                break;
            case CONFIG_TYPE_INT64:
                option = boost::shared_ptr<GenericOption> (new IntOption(setting->value.llval));
                break;
            case CONFIG_TYPE_STRING:
                option = boost::shared_ptr<GenericOption> (new Option(setting->value.sval));
                break;
            case CONFIG_TYPE_BOOL:
                option = boost::shared_ptr<GenericOption> (new BoolOption(setting->value.ival));
                break;
        }
    }
    config_destroy(&config);
    return option;
}

const void ConfigManager::getParameter(std::string path, std::string& value)
{
    boost::shared_ptr<Option> option(boost::dynamic_pointer_cast <Option> (readParameter(path)));
    if (NULL == option)
    {
        throw _Exception("Unknown parameter: " + path);
    }
    value = option->getOption();
    option.reset();
}

const void ConfigManager::getParameter(std::string path, int& value)
{
    boost::shared_ptr<IntOption> option(boost::dynamic_pointer_cast <IntOption> (readParameter(path)));
    if (NULL == option)
    {
        throw _Exception("Unknown parameter: " + path);
    }
    value = option->getIntOption();
    option.reset();
}

const void ConfigManager::getParameter(std::string path, bool& value)
{
    boost::shared_ptr<BoolOption> option (boost::dynamic_pointer_cast <BoolOption> (readParameter(path)));
    if (NULL == option)
    {
        throw _Exception("Unknown parameter:" + path);
    }
    value = option->getBoolOption();
    option.reset();
}

void ConfigManager::writeParameter(std::string path, boost::shared_ptr<GenericOption> value)
{
    config_t config;
    config_setting_t* setting = NULL;
    config_setting_t* parent;
    std::string parentPath, optionPath;

    config_init(&config);
    if (!config_read_file(&config, m_filename.c_str()))
    {
        std::string err = "Error in configuration: " + m_filename + ":" +
                          util::str(config_error_line(&config)) + " " +
                          config_error_text(&config);
        config_destroy(&config);
        throw _Exception(err);
    }

    std::list<std::string> pathElements;
    util::strsplit(path, '.', pathElements);

    parent = config_root_setting(&config);
    if (pathElements.size() > 1)
    {
        optionPath = pathElements.back();
        pathElements.pop_back();
        std::list<std::string>::iterator n = pathElements.begin();
        while (n != pathElements.end() && optionPath.compare(*n))
        {
            if (parentPath.size()) parentPath += ".";
            parentPath += *n;

            setting = config_lookup(&config, parentPath.c_str());
            if (NULL == setting)
            {
                setting = config_setting_add(parent, (*n).c_str(), CONFIG_TYPE_GROUP);
            }
            else if (! config_setting_is_group(setting))
            {
                config_destroy(&config);
                throw _Exception("Invalid parameter path: " + path);
            }
            parent = setting;
            setting = NULL;
            n++;
        }
    }
    else
    {
        parentPath = "";
        optionPath = path;
    }
    setting = config_lookup(&config, path.c_str());

    if (typeid(*value) == typeid(IntOption))
    {
        if (!setting)
        {
            setting = config_setting_add(parent, optionPath.c_str(), CONFIG_TYPE_INT);
        }
        if (!setting || config_setting_type(setting) != CONFIG_TYPE_INT)
        {
            config_destroy(&config);
            throw _Exception("Invalid parameter type, should be INT");
        }
        config_setting_set_int(setting, value->getIntOption());
    }
    else if (typeid(*value) == typeid(Option))
    {
        if (!setting)
        {
            setting = config_setting_add(parent, optionPath.c_str(), CONFIG_TYPE_STRING);
        }
        if (!setting || config_setting_type(setting) != CONFIG_TYPE_STRING)
        {
            config_destroy(&config);
            throw _Exception("Invalid parameter type, should be STRING");
        }
        config_setting_set_string(setting, value->getOption().c_str());
    }
    else if (typeid(*value) == typeid(BoolOption))
    {
        if (!setting)
        {
            setting = config_setting_add(parent, optionPath.c_str(), CONFIG_TYPE_BOOL);
        }
        if (!setting || config_setting_type(setting) != CONFIG_TYPE_BOOL)
        {
            config_destroy(&config);
            throw _Exception("Invalid parameter type, should be BOOL");
        }
        config_setting_set_bool(setting, value->getBoolOption());
    }
    else
    {
        config_destroy(&config);
        throw _Exception("Invalid configuration data type");
    }

    value.reset();
    config_write_file(&config, m_filename.c_str());
    config_destroy(&config);
}

const void ConfigManager::setParameter(std::string path, std::string value)
{
    writeParameter(path, boost::shared_ptr<Option>(new Option(value)));
}

const void ConfigManager::setParameter(std::string path, int value)
{
    writeParameter(path, boost::shared_ptr<IntOption>(new IntOption(value)));
}

const void ConfigManager::setParameter(std::string path, bool value)
{
    writeParameter(path, boost::shared_ptr<BoolOption>(new BoolOption(value)));
}

ConfigManager::~ConfigManager()
{
    for (size_t i = 0; i < m_options.size(); i++)
    {
        if (m_options[i] != NULL)
        {
            delete m_options[i];
        }
    }
    m_options.clear();
}

} // namespace

