/*
    Copyright (c) 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include <stdio.h>
#include <sstream>

#include "exceptions.h"
#include "database.h"
#include "util.h"
#include "dsmapi-wrapper.h"
#include "html_report_web_handler.h"
#include "gettext_helper.h"

#define MAX_FETCH_RECORDS                   50

namespace dSA
{

HTMLReportWebHandler::HTMLReportWebHandler(struct mg_connection* connection,
                               const struct mg_request_info* request_info) :
                               HTMLWebHandler(connection, request_info) {}


std::string HTMLReportWebHandler::taskOverview(
                                            boost::shared_ptr<dSLongTask> task)
{
    std::ostringstream result;

    result << header(_("Overview"));

    result << table_begin();

    result << table_row_begin();
    result << table_column_add(task->getName(), 2);
    result << table_row_end();

    result << table_row_begin();
    result << table_column_add(task->getDescription() + "<br/>", 2);
    result << table_row_end();

    result << table_row_begin();
    result << table_column_add(_("Status:"), 0, 250);
    result << table_column_add(task->getStatusInfo());
    result << table_row_end();

    task_status t = task->getTaskStatus();
    if (t == Queued)
    {
        result << table_end();
        return result.str();
    }

    result << table_row_simple(_("Started:"),
                            timeutil::epoch2local((time_t)task->getStartedTime()));

    if ((t != Cancelled) && (t != Running))
    {
        result << table_row_simple(_("Ended:"),
                            timeutil::epoch2local((time_t)task->getFinishedTime()));
    }

    if (t == Running)
    {
        result << table_row_begin();
        result << table_column_add(_("<b>Note:</b> a protocol can only be "
                                   "generated after the test has completed."),
                                   2);
        result << table_row_end();

        result << table_end();
        return result.str();
    }


    result << table_end();
    return result.str();
}

std::string HTMLReportWebHandler::upstreamPercent(int up)
{
    return util::str(up * 100 / TRANSMISSION_QUALITY_UPSTREAM_MAX) + '%';
}

std::string HTMLReportWebHandler::downstreamPercent(int down)
{
    return util::str(100 - (down * 100 / TRANSMISSION_QUALITY_DOWNSTREAM_MAX)) +
           '%';
}

std::string HTMLReportWebHandler::transmissionQualitySummary(long long int id)
{
    std::ostringstream result;

    boost::shared_ptr<Database::task_result_summary> summary =
        Database::getInstance()->getTaskResultSummary(id);

    boost::shared_ptr<Database::result_data> data = summary->data;

    result << header(_("Summary"));
    result << table_begin();

    result << table_row_begin();
    result << table_column_add(_("Number of tested devices:"), 0, 250);
    result << table_column_add((*data)["devices"]->first);
    result << table_row_end();

    result << table_row_simple(_("Number of executed tests:"),
                               (*data)["tests"]->first);

    result << table_row_simple(_("Number of tests with errors:"),
                               (*data)["errors"]->first);

    result << table_row_begin();
    result << table_column_add("&nbsp;", 2);
    result << table_row_end();


    result << table_row_simple(_("Average upstream quality:"),
                upstreamPercent(util::str2int((*data)["avgUpstream"]->first)));

    result << table_row_simple(_("Maximum upstream quality:"),
                upstreamPercent(util::str2int((*data)["maxUpstream"]->first)));

    result << table_row_simple(_("Minimum upstream quality:"),
                upstreamPercent(util::str2int((*data)["minUpstream"]->first)));

    result << table_row_begin();
    result << table_column_add("&nbsp;", 2);
    result << table_row_end();

    result << table_row_simple(_("Average downstream quality:"),
            downstreamPercent(util::str2int((*data)["avgDownstream"]->first)));

    result << table_row_simple(_("Maximum downstream quality:"),
            downstreamPercent(util::str2int((*data)["maxDownstream"]->first)));

    result << table_row_simple(_("Minimum downstream quality:"),
            downstreamPercent(util::str2int((*data)["minDownstream"]->first)));

    result << table_end();
    return result.str();
}

std::string HTMLReportWebHandler::renderTQDevREcord(
        boost::shared_ptr<
            const Database::transmission_quality_device_record> rec)
{
    std::ostringstream result;

    result << table_row_begin();
    result << table_column_add(rec->dsid);
    if (rec->errorcode == 0)
    {
        result << table_column_add(upstreamPercent(rec->upstream));
        result << table_column_add(downstreamPercent(rec->downstream));
    }
    else
    {
        result << table_column_add("-");
        result << table_column_add("-");
    }
    result << table_column_add(dSMApiWrapper::error2Message(rec->errorcode));
    result << table_column_add(timeutil::epoch2local(rec->timestamp));
    result << table_row_end();
    return result.str();
}

void HTMLReportWebHandler::transmissionQualityReport(
                                            boost::shared_ptr<dSLongTask> task,
                                            bool download)
{
    long long int c = 0;
    std::string s = page_begin(_("Protocol of transmission quality"), download);
    mg_write(m_connection, s.c_str(), s.length());

    try
    {
        s = taskOverview(task);
        mg_write(m_connection, s.c_str(), s.length());

        task_status t = task->getTaskStatus();
        if ((t != Finished) && (t != Cancelled))
        {
            s = page_end();
            mg_write(m_connection, s.c_str(), s.length());
            return;
        }

        s = transmissionQualitySummary(task->getTaskID());
        mg_write(m_connection, s.c_str(), s.length());

        Database* db = Database::getInstance();
        c = db->transmissionQualityResultCount(task->getTaskID());
        long long int offset = 0;

        std::ostringstream html;

        if (c > 0)
        {
            s = header(_("Detailed results"));
            mg_write(m_connection, s.c_str(), s.length());

            html << table_begin();

            html << table_row_begin();
            html << table_header(_("dSID"));
            html << table_header(_("Upstream"));
            html << table_header(_("Downstream"));
            html << table_header(_("Result"));
            html << table_header(_("Time"));
            html << table_row_end();
        }

        while (offset <= c)
        {
            boost::shared_ptr<
                std::list<boost::shared_ptr<
                    const Database::transmission_quality_device_record> > >
                        records = db->getTransmissionQualityResults(
                            task->getTaskID(), MAX_FETCH_RECORDS, offset);
            if (records->empty())
            {
                break;
            }

            while (!records->empty())
            {
                boost::shared_ptr<
                    const Database::transmission_quality_device_record> rec =
                        records->front();
                records->pop_front();
                html << renderTQDevREcord(rec);
            }
            mg_write(m_connection, html.str().c_str(), html.str().length());
            html.clear();
            html.str("");

            offset = offset + MAX_FETCH_RECORDS;
        }

        if (c > 0)
        {
            html << table_end();
            mg_write(m_connection, html.str().c_str(), html.str().length());
        }

    }
    catch (Exception &ex)
    {
        if (c > 0)
        {
            s = table_end();
            mg_write(m_connection, s.c_str(), s.length());
        }

        s = error(_("Could not complete report:") + std::string(" ") +
                  ex.getMessage());
        mg_write(m_connection, s.c_str(), s.length());
    }

    s = page_end();
    mg_write(m_connection, s.c_str(), s.length());
}

void HTMLReportWebHandler::testDevicePresenceReport(
        boost::shared_ptr<dSLongTask> task,
        bool download)
{
    long long int c = 0;
    std::string s = page_begin(_("Protocol of device presence"), download);
    mg_write(m_connection, s.c_str(), s.length());

    try
    {
        s = taskOverview(task);
        mg_write(m_connection, s.c_str(), s.length());

        task_status t = task->getTaskStatus();
        if ((t != Finished) && (t != Cancelled))
        {
            s = page_end();
            mg_write(m_connection, s.c_str(), s.length());
            return;
        }

        Database* db = Database::getInstance();
        boost::shared_ptr<TestDevicePresenceTask> dptask = boost::dynamic_pointer_cast<TestDevicePresenceTask> (task);
        webutil::query params;
        webutil::parse_query(dptask->getParameters(), params);
        dsid_t dsmdsid = util::str2dsid(params["dsm"]);

        // Summary
        boost::shared_ptr<Database::task_result_summary> summary =
                db->getTaskResultSummary(task->getTaskID());
        {
            std::ostringstream result;
            boost::shared_ptr<Database::result_data> data = summary->data;

            result << header(_("Summary"));
            result << table_begin();

            if (! IsNullId(dsmdsid))
            {
                result << table_row_begin();
                result << table_column_add(_("dSM-dSID of the power circuit:"),
                                           0, 250);
                result << table_column_add(util::str(dsmdsid));
                result << table_row_end();
            }
            result << table_row_begin();
            result << table_column_add(_("Number of tested devices:"), 0, 250);
            result << table_column_add((*data)["tests"]->first);
            result << table_row_end();

            c = util::str2llint((*data)["tests"]->first);
            long long int replies = util::str2llint((*data)["replies"]->first);
            long long int active = util::str2llint((*data)["active"]->first);

            result << table_row_simple(_("Number of newly detected devices:"),
                    (*data)["active"]->first);
            result << table_row_simple(_("Number of removed devices:"),
                    (*data)["inactive"]->first);
            result << table_row_simple(
                                _("Number of active devices after the test:"),
                                util::str(replies + active));

            result << table_row_begin();
            result << table_column_add("&nbsp;", 2);
            result << table_row_end();

            result << table_end();

            s = result.str();
            mg_write(m_connection, s.c_str(), s.length());
        }

        // Details
        long long int offset = 0;
        std::ostringstream html;

        if (c > 0)
        {
            s = header(_("Detailed results"));
            mg_write(m_connection, s.c_str(), s.length());

            html << table_begin();
            html << table_row_begin();
            if (IsNullId(dsmdsid))
            {
                html << table_header(_("dSM"));
            }
            html << table_header(_("Device-dSID"));
            html << table_header(_("Result"));
            html << table_header(_("Time"));
            html << table_row_end();
        }

        while (offset <= c)
        {
            boost::shared_ptr<
                std::list<boost::shared_ptr<
                    const Database::device_presence_record> > >
                        records = db->getDevicePresenceResults(
                            task->getTaskID(), MAX_FETCH_RECORDS, offset);
            if (records->empty())
            {
                break;
            }

            while (!records->empty())
            {
                boost::shared_ptr<
                    const Database::device_presence_record> rec =
                        records->front();
                records->pop_front();

                std::ostringstream result;
                result << table_row_begin();
                if (IsNullId(dsmdsid))
                    result << table_column_add(rec->dsmdid);
                result << table_column_add(rec->dsid.substr(rec->dsid.length() - 8, 8));
                switch(rec->resultcode)
                {
                    case 5:
                        result << table_column_add(
                                std::string(_("Inactive (was active before)")));
                        break;
                    case 6:
                        result << table_column_add(
                                std::string(_("Active (was inactive before)")));
                        break;
                    case 4:
                        result << table_column_add(std::string(_("Active")));
                        break;
                    case 3:
                        result << table_column_add(std::string(_("Inactive")));
                        break;
                    case 2:
                        result << table_column_add(
                                std::string(_("Not queried")));
                        break;
                }
                result << table_column_add(timeutil::epoch2local(rec->timestamp));
                result << table_row_end();
                html << result.str();
            }
            mg_write(m_connection, html.str().c_str(), html.str().length());
            html.clear();
            html.str("");

            offset = offset + MAX_FETCH_RECORDS;
        }

        if (c > 0)
        {
            html << table_end();
            mg_write(m_connection, html.str().c_str(), html.str().length());
        }

    }
    catch (Exception &ex)
    {
        if (c > 0)
        {
            s = table_end();
            mg_write(m_connection, s.c_str(), s.length());
        }

        s = error(_("Could not complete report: ") + ex.getMessage());
        mg_write(m_connection, s.c_str(), s.length());
    }

    s = page_end();
    mg_write(m_connection, s.c_str(), s.length());
}

void HTMLReportWebHandler::taskReport()
{
    std::string task_id = m_parameters["id"];
    if (task_id.empty())
    {
        failure(_("Error"),
                _("Can not generate report: missing task id parameter!"));
        return;
    }

    bool dl = false;
    std::string download = m_parameters["download"];
    if (!download.empty())
    {
        try
        {
            dl = (util::str2int(download));
        }
        catch (Exception &ex)
        {
            failure(_("Error"),
                    _("Can not generate report: invalid download parameter!"));
            return;
        }
    }
    boost::shared_ptr<dSLongTask> task;
    try
    {
        task = Database::getInstance()->getTask(util::str2llint(task_id));
    }
    catch (Exception &ex)
    {
        failure(_("Error"), _("Can not generate report!") + std::string(" ") +
                ex.getMessage());
        return;
    }

    switch (task->getTaskType())
    {
        case TestTransmissionQuality:
            transmissionQualityReport(task, dl);
            break;
        case TestDevicePresence:
            testDevicePresenceReport(task, dl);
            break;
        default:
            failure(_("Error"),
                    _("Can not generate report: unknown task type!"));
            return;
    }
}

void HTMLReportWebHandler::process()
{
    // unlike the json calls the HTML report can probably get quite big,
    // therefore we will try to keep a lesser amount of data in memory and
    // avoid passing large strings around, instead each report has to
    // write chunks of data using mg_write()
    try
    {
        if (m_command == "taskReport")
        {
            taskReport();
        }
        else
        {
            failure(_("Error - can not generate report!"),
                    _("Unknown function called"));
        }
    }
    // we can't return anything to the user at this point because data is
    // written out in chunks, meaning that the user received at least a part of
    // the page already; the functions that take care of processing each call
    // have to ensure that the html never gets broken and thus handle all
    // possible exceptions gracefully
    catch (Exception &ex)
    {
        fprintf(stderr, "Error generating report: %s\n",
                ex.getMessage().c_str());
    }
    catch (...)
    {
        fprintf(stderr, "Encountered unexpected error while processing call");
    }
}

} // namespace

