/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include <boost/shared_ptr.hpp>
#include <autosprintf.h>

#include "json_hardware_web_handler.h"
#include "storage.h"
#include "config_manager.h"
#include "ds-assistant.h"
#include "exceptions.h"
#include "json_util.h"
#include "gettext_helper.h"

#define DEFAULT_CHECKS_PER_DEVICE 1

using gnu::autosprintf;

namespace dSA
{

JSONHardwareWebHandler::JSONHardwareWebHandler(struct mg_connection* connection,
                               const struct mg_request_info* request_info) :
                               JSONWebHandler(connection, request_info) {}

std::string JSONHardwareWebHandler::get()
{
    std::ostringstream result;
    boost::shared_ptr<std::list<dsid_t> > dsm_list;

    Storage *storage = Storage::getInstance();

    std::string dsid = m_parameters["dsm"];
    if (dsid.empty())
    {
        dsm_list = storage->getDSMList();
    }
    else
    {
        dsid_t dsm_dsid = util::str2dsid(dsid);
        dsm_list = boost::shared_ptr<std::list<dsid_t> >(new std::list<dsid_t>);
        dsm_list->push_back(dsm_dsid);
    }

    json_object *ret = json_ok();
    json_object *data = json_object_new_object();
    json_object *dsms = json_object_new_array();
    json_object_object_add(data, "dsms", dsms);

    while (!dsm_list->empty())
    {
        boost::shared_ptr<const dSM> dsm = storage->getDSM(dsm_list->front());
        dsm_list->pop_front();

        if (dsm == NULL)
        {
            continue;
        }

        json_object *jdsm = json_object_new_object();

        json_object_object_add(jdsm, "dsid",
                json_object_new_string(util::str(dsm->getDSID()).c_str()));
        json_object_object_add(jdsm, "hardware",
                json_object_new_int(dsm->getHWVersion()));
        json_object_object_add(jdsm, "arm",
                json_object_new_int(dsm->getARMSwVersion()));
        json_object_object_add(jdsm, "dsp",
                json_object_new_int(dsm->getDSPSwVersion()));
        json_object_object_add(jdsm, "api",
                json_object_new_int(dsm->getAPIVersion()));
        json_object_object_add(jdsm, "name",
                json_object_new_string(dsm->getName().c_str()));
        json_object_object_add(jdsm, "present",
                json_object_new_boolean((bool)dsm->getPresent()));

        boost::shared_ptr<std::list<boost::shared_ptr<Device> > > devs =
            storage->getDevices(dsm->getDSID());

        json_object *devices = json_object_new_array();

        while (!devs->empty())
        {
            boost::shared_ptr<Device> device = devs->front();
            devs->pop_front();

            if (device == NULL)
            {
                continue;
            }

            json_object_array_add(devices, jsonutil::device2json(device));
        }

        json_object_object_add(jdsm, "devices", devices);

        json_object_array_add(dsms, jdsm);
    }

    json_object_object_add(ret, "data", data);

    result << json_object_to_json_string(ret);

    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::getDevice()
{
    std::ostringstream result;
    boost::shared_ptr<Device> device;

    Storage *storage = Storage::getInstance();

    std::string device_dsid = m_parameters["dsid"];
    if (device_dsid.empty())
    {
        throw _Exception("Parameter dsid is missing");
    }

    std::string dsm_dsid = m_parameters["dsm"];
    if (dsm_dsid.empty())
    {
        device = storage->getDevice(util::str2dsid(device_dsid));
    }
    else
    {
        device = storage->getDevice(util::str2dsid(device_dsid),
                                    util::str2dsid(dsm_dsid));
    }


    json_object *ret = json_ok();
    json_object *data = json_object_new_object();
    json_object_object_add(data, "device", jsonutil::device2json(device));
    json_object_object_add(ret, "data", data);

    result << json_object_to_json_string(ret);

    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::getApartmentId()
{
    std::ostringstream result;
    int appid;
    ConfigManager::getInstance()->getParameter("digitalstrom.apartmentid", appid);
    dsid_t device_list[DSM_API_MAX_DEVICES];
    int busmembercount = dS_Assistant::getInstance()->dS485()->getBusMembers(
                                            device_list, DSM_API_MAX_DEVICES);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();
    json_object_object_add(res, "apartmentid", json_object_new_int(appid));
    json_object_object_add(res, "busmembercount",
                                        json_object_new_int(busmembercount));
    json_object_object_add(ret, "data", res);
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::setApartmentId()
{
    std::ostringstream result;
    dsid_t device_list[DSM_API_MAX_DEVICES];
    int busmembercount = dS_Assistant::getInstance()->dS485()->getBusMembers(
                                            device_list, DSM_API_MAX_DEVICES);

    std::string aids = m_parameters["id"];
    if (aids.empty())
    {
        throw _Exception(_("parameter id is missing"));
    }

    int aid = util::str2int(aids);
    if (aid < 0 || aid + busmembercount >= 63)
    {
        throw _Exception(_("Parameter id has invalid value"));
    }

    // save value to database
    ConfigManager::getInstance()->setParameter("digitalstrom.apartmentid", aid);

    // save value to devices
    dsid_t alldsmid;
    SetNullId(alldsmid);
    boost::shared_ptr<SetDSMApartmentIdTask> task(
                                new SetDSMApartmentIdTask(alldsmid, aid, true));
    dS_Assistant::getInstance()->addTask(
                                    boost::dynamic_pointer_cast<dSTask>(task));

    json_object *ret = json_ok();
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::getMeterProgrammingModeFlag()
{
    std::ostringstream result;
    bool pmode;
    ConfigManager::getInstance()->getParameter("digitalstrom.programmingmode", pmode);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();
    json_object_object_add(res, "progmode", json_object_new_boolean(pmode));
    json_object_object_add(ret, "data", res);
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::setMeterProgrammingModeFlag()
{
    std::ostringstream result;

    std::string pmodestring = m_parameters["mode"];
    if (pmodestring.empty())
    {
        throw _Exception(_("parameter mode is missing"));
    }

    bool pmode = true;
    bool dbmode;
    ConfigManager::getInstance()->getParameter(
            "digitalstrom.programmingmode", dbmode);

    if (pmodestring.compare("disable") == 0)
    {
        pmode = false;
    }
    else if (pmodestring.compare("enable") == 0)
    {
        pmode = true;
    }
    else
    {
        throw _Exception(_("parameter mode has invalid value") +
                         std::string(": ") + pmodestring);
    }

    if (dbmode != pmode)
    {
        // save value to database
        ConfigManager::getInstance()->setParameter(
                "digitalstrom.programmingmode", pmode);

        // save value to devices
        boost::shared_ptr<CheckDSMThread> task(new CheckDSMThread(0));
        dS_Assistant::getInstance()->addLongTask(
                                boost::dynamic_pointer_cast<dSLongTask>(task));
    }

    json_object *ret = json_ok();
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::getMeterRoomDimmingModeFlag()
{
    std::ostringstream result;
    bool pmode;
    ConfigManager::getInstance()->getParameter("digitalstrom.roomdimmingmode", pmode);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();
    json_object_object_add(res, "roomdimmmode", json_object_new_boolean(pmode));
    json_object_object_add(ret, "data", res);
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::setMeterRoomDimmingModeFlag()
{
    std::ostringstream result;

    std::string rdmodestring = m_parameters["mode"];
    if (rdmodestring.empty())
    {
        throw _Exception("Parameter mode is missing");
    }

    bool rdmode = true;
    bool dbmode;
    ConfigManager::getInstance()->getParameter(
            "digitalstrom.roomdimmingmode", dbmode);

    if (rdmodestring.compare("disable") == 0)
    {
        rdmode = false;
    }
    else if (rdmodestring.compare("enable") == 0)
    {
        rdmode = true;
    }
    else
    {
        throw _Exception("Parameter mode has invalid value: " + rdmodestring);
    }

    if (dbmode != rdmode)
    {
        // save value to database
        ConfigManager::getInstance()->setParameter(
                "digitalstrom.roomdimmingmode", rdmode);

        // save value to devices
        boost::shared_ptr<CheckDSMThread> task(new CheckDSMThread(0));
        dS_Assistant::getInstance()->addLongTask(
                                boost::dynamic_pointer_cast<dSLongTask>(task));
    }

    json_object *ret = json_ok();
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::getMeterLedModeFlag()
{
    std::ostringstream result;
    bool ledmode;
    ConfigManager::getInstance()->getParameter("digitalstrom.meter.led",
            ledmode);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();
    json_object_object_add(res, "ledmode", json_object_new_boolean(ledmode));
    json_object_object_add(ret, "data", res);
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::setMeterLedModeFlag()
{
    std::ostringstream result;

    std::string lmodestring = m_parameters["mode"];
    if (lmodestring.empty())
    {
        throw _Exception(_("parameter mode is missing"));
    }

    bool ledmode = true;
    bool dbmode;
    ConfigManager::getInstance()->getParameter(
                "digitalstrom.meter.led", dbmode);
    if (lmodestring.compare("disable") == 0)
    {
        ledmode = false;
    }
    else if (lmodestring.compare("enable") == 0 )
    {
        ledmode = true;
    }
    else
    {
        throw _Exception(autosprintf(_("parameter mode has invalid value: %s"),
                         lmodestring.c_str()));
    }

    if (ledmode != dbmode)
    {
        // save value to database
        ConfigManager::getInstance()->setParameter(
                "digitalstrom.meter.led", ledmode);

        // save value to devices
        boost::shared_ptr<CheckDSMThread> task(new CheckDSMThread(0));
        dS_Assistant::getInstance()->addLongTask(
                                boost::dynamic_pointer_cast<dSLongTask>(task));
    }

    json_object *ret = json_ok();
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::testDeviceTransmissionQuality()
{
    std::ostringstream result;

    int checksPerDevice = getOptionalIntParam(m_parameters["checksPerDevice"],
                                              DEFAULT_CHECKS_PER_DEVICE,
                                              "checksPerDevice");

    if (checksPerDevice < 1)
    {
        throw _Exception(autosprintf(
                    _("parameter checksPerDevice has invalid value: %s"),
                       util::str(checksPerDevice).c_str()));
    }

    dsid_t dsm_dsid;

    std::string dsid = m_parameters["dsm"];
    if (!dsid.empty())
    {
        dsm_dsid = util::str2dsid(dsid);
        boost::shared_ptr<const dSM> dsm =
                                    Storage::getInstance()->getDSM(dsm_dsid);
        if (dsm == NULL)
        {
            throw _Exception(autosprintf(_("dSM %s not found"),
                                        util::str(dsm_dsid).c_str()));
        }
    }
    else
    {
        SetNullId(dsm_dsid);
    }

    dS_Assistant *dsa = dS_Assistant::getInstance();

    boost::shared_ptr<TestTranmissionQualityTask> task(
            new TestTranmissionQualityTask(checksPerDevice, dsm_dsid));
    task->setPriority(High);
    unsigned long task_id = dsa->addLongTask(task);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();

    json_object_object_add(res, "task",
            json_object_new_string(util::str(task_id).c_str()));
    json_object_object_add(ret, "data", res);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::testDevicePresence()
{
    std::ostringstream result;

    int interval_time = getOptionalIntParam(m_parameters["interval"],
                                              2000,
                                              "interval");
    dsid_t dsm_dsid;

    std::string dsid = m_parameters["dsm"];
    if (!dsid.empty())
    {
        dsm_dsid = util::str2dsid(dsid);
        boost::shared_ptr<dSM> dsm = Storage::getInstance()->getDSM(dsm_dsid);

        uint8_t Hardware;
        uint8_t DSP;
        uint8_t Temperature;
        uint8_t Config;
        uint8_t dS485;
        uint8_t Line;
        dS_Assistant::getInstance()->dS485()->dSMState(dsm_dsid,
                &Hardware, &DSP, &Temperature, &Config, &dS485, &Line);
        if (Line)
        {
            throw _Exception(autosprintf(
                _("dSM %s is currently registering devices. Please run the "
                  "test again later"), util::str(dsm_dsid).c_str()));
        }
    }
    else
    {
        SetNullId(dsm_dsid);
    }

    dS_Assistant *dsa = dS_Assistant::getInstance();

    boost::shared_ptr<TestDevicePresenceTask> task(
            new TestDevicePresenceTask(dsm_dsid, interval_time));

    task->setPriority(High);
    unsigned long task_id = dsa->addLongTask(task);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();

    json_object_object_add(res, "task",
            json_object_new_string(util::str(task_id).c_str()));
    json_object_object_add(ret, "data", res);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::testDeviceStatistic(int mode)
{
    std::ostringstream result;
    dsid_t dsm_dsid;
    dsid_t dev_dsid;

    std::string dsmid = m_parameters["dsm"];
    std::string devid = m_parameters["dsid"];

    SetNullId(dsm_dsid);
    SetNullId(dev_dsid);

    if (!dsmid.empty())
    {
        dsm_dsid = util::str2dsid(dsmid);
    }
    if (!devid.empty())
    {
        dev_dsid = util::str2dsid(devid);
    }

    dS_Assistant *dsa = dS_Assistant::getInstance();
    boost::shared_ptr<dSLongTask> task;

    switch (mode)
    {
        case 0:
        default:
            task = boost::shared_ptr<dSLongTask>
                (new TestDeviceStatisticTask(dsm_dsid, dev_dsid));
            break;
        case 1:
            task = boost::shared_ptr<dSLongTask>
                (new TestDeviceStatisticResetTask(dsm_dsid, dev_dsid));
            break;
    }

    task->setPriority(High);
    unsigned long task_id = dsa->addLongTask(task);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();

    json_object_object_add(res, "task",
            json_object_new_string(util::str(task_id).c_str()));
    json_object_object_add(ret, "data", res);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::testMeterStatistic(int mode)
{
    std::ostringstream result;
    dsid_t dsm_dsid;

    std::string dsid = m_parameters["dsm"];
    if (!dsid.empty())
    {
        dsm_dsid = util::str2dsid(dsid);
    }
    else
    {
        SetNullId(dsm_dsid);
    }

    dS_Assistant *dsa = dS_Assistant::getInstance();
    boost::shared_ptr<dSLongTask> task;

    switch (mode)
    {
        case 0:
        default:
            task = boost::shared_ptr<dSLongTask>
                (new TestMeterStatisticTask(dsm_dsid));
            break;
        case 1:
            task = boost::shared_ptr<dSLongTask>
                (new TestMeterStatisticResetTask(dsm_dsid));
            break;
    }

    task->setPriority(High);
    unsigned long task_id = dsa->addLongTask(task);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();

    json_object_object_add(res, "task",
            json_object_new_string(util::str(task_id).c_str()));
    json_object_object_add(ret, "data", res);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::testSystemMonitor()
{
    std::ostringstream result;
    int profile = getOptionalIntParam(m_parameters["profile"], 1, "profile");
    int duration = getOptionalIntParam(m_parameters["duration"], 3600, "duration");

    dS_Assistant *dsa = dS_Assistant::getInstance();

    // TODO: there should be a single system test running at a time!

    boost::shared_ptr<TestSystemMonitorTask> task(
            new TestSystemMonitorTask(profile, duration));

    task->setPriority(Normal);
    unsigned long task_id = dsa->addLongTask(task);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();

    json_object_object_add(res, "task",
            json_object_new_string(util::str(task_id).c_str()));
    json_object_object_add(ret, "data", res);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::testDeviceMonitor()
{
    std::ostringstream result;
    dS_Assistant *dsa = dS_Assistant::getInstance();

    boost::shared_ptr<DeviceMonitorTask> task(
            new DeviceMonitorTask());

    task->setPriority(Normal);
    unsigned long task_id = dsa->addLongTask(task);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();

    json_object_object_add(res, "task",
            json_object_new_string(util::str(task_id).c_str()));
    json_object_object_add(ret, "data", res);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONHardwareWebHandler::testDeviceMonitorStatus()
{
    std::ostringstream result;
    dS_Assistant *dsa = dS_Assistant::getInstance();

    std::string task_id = m_parameters["id"];
    if (task_id.empty())
    {
        throw _Exception(_("missing task id parameter"));
    }
    long long int tid = util::str2llint(task_id);

    boost::shared_ptr<DeviceMonitorTask> task =
        boost::dynamic_pointer_cast<DeviceMonitorTask> (dsa->getTaskInfo(tid));

    if (!task)
    {
        throw _Exception("device monitor task not found");
    }

    json_object *ret = json_ok();
    json_object *nfo;

    if (task != NULL)
    {
        nfo = json_object_new_object();
        json_object_object_add(nfo, "id",
            json_object_new_string(util::str(task->getTaskID()).c_str()));
        json_object_object_add(nfo, "name",
            json_object_new_string(task->getName().c_str()));
        json_object_object_add(nfo, "description",
            json_object_new_string(task->getDescription().c_str()));
        json_object_object_add(nfo, "status",
            json_object_new_string(task->getTaskStatusText().c_str()));
        json_object_object_add(nfo, "statusInfo",
            json_object_new_string(task->getStatusInfo().c_str()));
        json_object_object_add(nfo, "type",
            json_object_new_int((int) task->getTaskType()));
        json_object_object_add(nfo, "device",
            json_object_new_string(task->getLastDSID().c_str()));
        json_object_object_add(nfo, "dsmeter",
            json_object_new_string(task->getLastDSMDSID().c_str()));
    }
    else
    {
        nfo = json_object_new_object();
        json_object_object_add(nfo, "id",
            json_object_new_string(util::str(dsa->getInvalidTaskID()).c_str()));
    }

    json_object_object_add(ret, "data", nfo);
    result << json_object_to_json_string(ret);
    json_object_put(ret);

    return result.str();
}

void JSONHardwareWebHandler::process()
{
    try
    {
        std::string result;
        std::string header = http_ok();

        if (m_command == "get")
        {
            result = get();
        }
        else if (m_command == "getDevice")
        {
            result = getDevice();
        }
        else if (m_command == "getApartmentId")
        {
            result = getApartmentId();
        }
        else if (m_command == "setApartmentId")
        {
            result = setApartmentId();
        }
        else if (m_command == "getMeterProgrammingModeFlag")
        {
            result = getMeterProgrammingModeFlag();
        }
        else if (m_command == "setMeterProgrammingModeFlag")
        {
            result = setMeterProgrammingModeFlag();
        }
        else if (m_command == "getMeterRoomDimmingModeFlag")
        {
            result = getMeterRoomDimmingModeFlag();
        }
        else if (m_command == "setMeterRoomDimmingModeFlag")
        {
            result = setMeterRoomDimmingModeFlag();
        }
        else if (m_command == "getMeterLedModeFlag")
        {
            result = getMeterLedModeFlag();
        }
        else if (m_command == "setMeterLedModeFlag")
        {
            result = setMeterLedModeFlag();
        }
        else if (m_command == "testDeviceTransmissionQuality")
        {
            result = testDeviceTransmissionQuality();
        }
        else if (m_command == "testDevicePresence")
        {
            result = testDevicePresence();
        }
        else if (m_command == "testDeviceStatistic")
        {
            result = testDeviceStatistic(0);
        }
        else if (m_command == "testMeterStatistic")
        {
            result = testMeterStatistic(0);
        }
        else if (m_command == "testDeviceStatisticReset")
        {
            result = testDeviceStatistic(1);
        }
        else if (m_command == "testMeterStatisticReset")
        {
            result = testMeterStatistic(1);
        }
        else if (m_command == "testSystemMonitor")
        {
            result = testSystemMonitor();
        }
        else if (m_command == "testDeviceMonitor")
        {
            result = testDeviceMonitor();
        }
        else if (m_command == "testDeviceMonitorStatus")
        {
            result = testDeviceMonitorStatus();
        }
        else
        {
            failure(_("unknown function called"));
            return;
        }

        mg_write(m_connection, header.c_str(), header.length());
        mg_write(m_connection, result.c_str(), result.length());
    }
    catch (Exception &ex)
    {
        failure(ex.getMessage().c_str());
    }
    catch (...)
    {
        failure(_("exception raised while processing call"));
    }
}

} // namespace

